import graphlib
import logging
import zipfile
from contextlib import contextmanager
from django.apps.registry import apps
from django.db.models import ManyToOneRel, Model
from io import TextIOWrapper, BufferedReader
from typing import Any, BinaryIO, Dict, Iterable, List, Optional, Set, Tuple, Type

from ..errors import ModelClassNotRegistered
from ..policy import ImportPolicy
from ..serializers import Exporter
from ..types import ID, Ref, ObjectData, Attachment
from .._util import UncloseableStream, get_model_options

from .base import BaseContainer
from ._yaml import get_yaml


logger = logging.getLogger('haul.import')
logger.setLevel(logging.DEBUG)


class ImportReport:
    loaded_objects: Set[ObjectData]
    imported_objects: Set[Model]
    discarded_objects: Set[ObjectData]
    pk_map: Dict[ID, Model]

    def __init__(self):
        self.loaded_objects = set()
        self.imported_objects = set()
        self.discarded_objects = set()
        self.pk_map = {}


class ImportContainer(BaseContainer):
    '''
    Your starting point for object import.
    '''
    __instance_map: Dict[ID, Model]
    __discarded_objects: Set[ID]

    #: free-form metadata as stored by :func:`ExportContainer.write`
    metadata: Any = None

    def __init__(
        self,
        exporters: List[Type[Exporter]] = [],
        policy: Optional[ImportPolicy] = None,
        ignore_unknown=False,
    ):
        super().__init__(exporters, ignore_unknown)
        self.__instance_map = {}
        self.__discarded_objects = set()
        self.__open = False
        self.policy = policy or ImportPolicy()
        self.report = ImportReport()

    @contextmanager
    def read(self, stream: BinaryIO):
        '''
        Reads a data stream, deserializes objects in it and stores them inside the container.

        This is a context manager which has to be kept open when :func:`import_objects` is called::

            c = ImportContainer(exporters=...)
            with open(...) as f:
                with c.read(f):
                    c.import_objects()
        '''
        stream.seek(0)
        stream = UncloseableStream(stream)
        reader = BufferedReader(stream)  # type: ignore
        signature = reader.peek(4)

        archive: Optional[zipfile.ZipFile] = None
        if signature[:2] == b'PK':
            logger.debug('Detected a ZIP container')
            archive = zipfile.ZipFile(reader, 'r')
            try:
                metadata_stream = archive.open('metadata.yaml', 'r')
            finally:
                archive.close()
        else:
            metadata_stream = reader

        try:
            try:
                all_kinds = set(ID.kind_for_model(x) for x in apps.get_models())
                yaml = get_yaml()
                for document in yaml.load_all(TextIOWrapper(metadata_stream)):
                    if document['_'] == 'header':
                        if document['version'] != 1:
                            raise ValueError(f'Unknown container version {document["version"]}')
                        unknown_kinds = set(document['object_kinds']) - all_kinds
                        if unknown_kinds and not self.ignore_unknown:
                            raise ValueError(f'Unknown object types {unknown_kinds}')
                        self.metadata = document.get('metadata')
                        if self.metadata:
                            logger.debug(f'Container metadata: {self.metadata}')
                    elif document['_'] == 'object':
                        id = document['id']
                        logger.debug(f'Extracting object {id}')
                        obj = ObjectData(
                            id=document['id'],
                            serialized_data=document['data'],
                            attachments=[
                                Attachment(
                                    id=item['id'],
                                    key=item['key'],
                                    _container_stream=stream,  # type: ignore
                                )
                                for item in document['attachments']
                            ]
                        )
                        if obj.id in self._objects:
                            raise ValueError(f'Duplicate object {obj.id} found')
                        self._objects[obj.id] = obj
                        self.report.loaded_objects.add(obj)
                    else:
                        raise ValueError(f'Unknown container segment "{document["_"]}"')
            finally:
                metadata_stream.close()

            try:
                self.__open = True
                yield
            finally:
                self.__open = False

        finally:
            if archive:
                archive.close()

    def _register_imported(self, obj: ObjectData, instance: Model):
        logger.debug(f'Imported {instance}')
        self.report.imported_objects.add(instance)
        self.__instance_map[obj.id] = instance

    def _discard_objects(self, objects: Iterable[ObjectData], reason=None):
        for kind, objects in self.__group_by_kind(objects).items():
            if len(objects) <= 5:
                description = ', '.join(str(x) for x in objects)
            else:
                description = f'{len(objects)} {kind} objects'
            logger.debug(f'Discarding {description} {reason or ""}')
            self.__discarded_objects |= {x.id for x in objects}
            self.report.discarded_objects |= set(objects)

    def import_objects(self) -> ImportReport:
        '''
        Untangles the object graph, relinks objects and imports them into the database.
        '''
        if not self.__open:
            raise RuntimeError('Container is not open - open a .read() context first')

        kind_map = self.__group_by_kind(self._objects.values())

        # ----------------
        # Deserialize data

        for kind, objects in list(kind_map.items()):
            try:
                exporter_cls = self._exporter_for_kind(kind)
            except ModelClassNotRegistered:
                if self.ignore_unknown:
                    self._discard_objects(objects, reason='due to unknown type')
                    continue
                raise

            exporter = exporter_cls(data=[x.serialized_data for x in objects], many=True)
            exporter.is_valid(raise_exception=True)
            logger.debug(f'Deserialized {len(objects)} {kind} objects')

            if len(objects) != len(exporter.validated_data):
                raise ValueError('Serializer has failed to deserialize all objects')

            for obj, deserialized_data in zip(objects, exporter.validated_data):
                obj.fields = deserialized_data
                assert obj.fields is not None
                for key, value in obj.fields.items():
                    # Foreign key
                    if isinstance(value, Ref):
                        obj.add_reference(value)
                        if len(value.ids):
                            logger.debug(f'Found a reference from {obj.id} to {value.ids}')

            for obj in self._objects.values():
                for ref in obj.refs:
                    for id in ref.ids:
                        if id not in self._objects:
                            raise ValueError(f'Unresolved reference to {id} from {obj.id} via {ref.field}')

        # -------------------
        # Build object graph

        sorter: graphlib.TopologicalSorter[ObjectData] = graphlib.TopologicalSorter(None)
        for obj in self._objects.values():
            if obj.id not in self.__discarded_objects:
                deps = [
                    self._objects[id]
                    for ref in obj.refs
                    for id in ref.ids
                    if id not in self.__discarded_objects and not ref.weak
                ]
                sorter.add(
                    obj,
                    *deps,
                )

        try:
            sorter.prepare()
        except graphlib.CycleError as e:
            logger.error('Cycle detected')
            for obj in e.args[1]:
                logger.error(f' - {obj}')
            raise e

        while sorter.is_active():
            ready_objects: Tuple[ObjectData, ...] = sorter.get_ready()
            if not len(ready_objects):
                raise RuntimeError('Could not untangle the reference graph')

            kind_map = self.__group_by_kind(ready_objects)

            # -------------------
            # Resolve all references

            for kind, objects in kind_map.items():
                model_meta = get_model_options(ID.model_for_kind(kind))

                for obj in objects:
                    assert obj.fields is not None
                    for ref in obj.refs:
                        if ref.weak:
                            continue
                        remaining_ids = list(ref.ids)
                        discarded = False

                        for id in ref.ids:
                            if id not in self.__instance_map:
                                if id in self.__discarded_objects:
                                    if ref.nullable:
                                        logger.debug(f'Breaking reference {obj.id}.{ref.field} due to target object being discarded')
                                        remaining_ids.remove(id)
                                        continue
                                    else:
                                        self._discard_objects([obj], reason=f'due to a broken reference via {ref.field}')
                                        sorter.done(obj)
                                        discarded = True
                                        break
                                raise ValueError(f'Consistency error: PK still unknown for {id} (referenced by {obj.id} via {ref.field})')

                        if discarded:
                            continue

                        if model_meta.get_field(ref.field).many_to_many:
                            obj.fields[ref.field] = [
                                self.__instance_map[id]
                                for id in remaining_ids
                            ]
                            if len(remaining_ids):
                                logger.debug(f'Remapped M2M {obj.id}.{ref.field} reference from {ref.ids} to {obj.fields[ref.field]}')
                        else:
                            if len(remaining_ids):
                                obj.fields[ref.field] = self.__instance_map[remaining_ids[0]]
                                logger.debug(f'Remapped {obj.id}.{ref.field} reference from {ref.ids[0]} to {obj.fields[ref.field].pk}')
                            else:
                                obj.fields[ref.field] = None

                    # Remove reverse FK fields
                    for relation in model_meta.related_objects:
                        if isinstance(relation, ManyToOneRel) and relation.related_name:
                            obj.fields.pop(relation.related_name, None)

                # -------------------
                # Gather relink actions

                for obj in objects:
                    assert obj.fields is not None
                    self.policy.preprocess_object_fields(ID.model_for_kind(kind), obj.fields)

                relink_actions = [
                    self.policy.relink_object(
                        ID.model_for_kind(kind),
                        obj,
                    )
                    for obj in objects
                    if obj.id not in self.__discarded_objects
                ]

                # -------------------
                # Execute relink actions

                for action in set(relink_actions):
                    action_objects = [x[1] for x in zip(relink_actions, objects) if x[0] == action]

                    for obj in action_objects:
                        assert obj.fields is not None
                        self.policy.postprocess_object_fields(ID.model_for_kind(kind), obj.fields)

                    logger.debug(f'Running {action} on {len(action_objects)} {kind} objects')
                    instances = action._execute(ID.model_for_kind(kind), action_objects, self.policy)
                    for obj, instance in zip(action_objects, instances):
                        if instance is False:
                            self._discard_objects([obj], reason='due to relink policy')
                        else:
                            self.policy.post_object_import(instance)
                            self._register_imported(obj, instance)
                        sorter.done(obj)

        # -------------------
        # Process attachments

        container_streams = set()
        for obj in self._objects.values():
            for attachment in obj.attachments:
                if not attachment._container_stream:
                    raise RuntimeError(f'Container stream not set for attachment {attachment}')
                container_streams.add(attachment._container_stream)

        for stream in container_streams:
            with zipfile.ZipFile(stream, 'r') as zfile:
                for obj in self._objects.values():
                    if obj.id in self.__discarded_objects:
                        continue
                    for attachment in obj.attachments:
                        if attachment._container_stream == stream:
                            with zfile.open(f'attachments/{attachment.id}', 'r') as f:
                                self.policy.process_attachment(self.__instance_map[obj.id], attachment.key, f)

        self.report.pk_map = self.__instance_map
        return self.report

    def __group_by_kind(self, objects: Iterable[ObjectData]) -> Dict[str, List[ObjectData]]:
        kind_map: Dict[str, List[ObjectData]] = {}
        for obj in objects:
            kind_map.setdefault(obj.id.kind, []).append(obj)
        return kind_map



'''
Haul uses DRF ``ModelSerializer`` as a base for its serialization.

This means you can use custom ``Field`` and ``Serializer`` objects to serialize and deserialize fields.
'''

from rest_framework import serializers
from rest_framework.validators import UniqueValidator

from .policy import ExportPolicy
from .types import ID, Ref


class DummyField(serializers.Field):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, read_only=True, allow_null=True, required=False)

    def run_validation(self, data=None):
        return None

    def get_attribute(self, instance):
        return None

    def to_representation(self, value):
        return None

    def to_internal_value(self, data):
        return None


class ForeignKey(serializers.Field):
    export_policy: ExportPolicy

    def __init__(self, *args, queryset=None, many=False, **kwargs):
        super().__init__(*args, read_only=False, **kwargs)
        self._many = many

    def bind(self, field_name, parent):
        super().bind(field_name, parent)
        self.export_policy = self.context.get('export_policy')

    def get_attribute(self, instance):
        return getattr(instance, self.source)

    def to_representation(self, value):
        if self._many:
            return [
                ID.from_object(instance)
                for instance in value
                if self.export_policy.should_export_object(instance)
            ]
        return ID.from_object(value) if self.export_policy.should_export_object(value) else None

    def to_internal_value(self, data):
        if self._many:
            return Ref(ids=data, field=self.source, nullable=self.allow_null)
        else:
            return Ref(ids=[data], field=self.source, nullable=self.allow_null)


class _BaseM2X(serializers.Field):
    weak = False
    export_policy: ExportPolicy

    def bind(self, field_name, parent):
        super().bind(field_name, parent)
        self.export_policy = self.context.get('export_policy')

    def get_attribute(self, instance):
        return getattr(instance, self.source).all()

    def to_representation(self, value):
        return [
            ID.from_object(instance)
            for instance in value
            if self.export_policy.should_export_object(instance)
        ]

    def to_internal_value(self, data):
        return Ref(data, field=self.source, nullable=True, weak=self.weak)


class ManyToMany(_BaseM2X):
    weak = False


class ReverseForeignKey(_BaseM2X):
    weak = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, allow_null=True)


class Exporter(serializers.ModelSerializer):
    serializer_related_field = DummyField

    def build_standard_field(self, field_name, model_field):
        cls, kwargs = super().build_standard_field(field_name, model_field)
        if 'validators' in kwargs:
            kwargs['validators'] = [x for x in kwargs['validators'] if not isinstance(x, UniqueValidator)]
        kwargs.pop('read_only', None)
        kwargs['required'] = False
        return cls, kwargs



from django.db.models import Model
from typing import Dict, List, Type

from ..errors import ModelClassNotRegistered
from ..serializers import Exporter
from ..types import ID, ObjectData


class BaseContainer:
    _objects: Dict[ID, ObjectData]

    def __init__(self, exporters: List[Type[Exporter]] = [], ignore_unknown=False):
        self.__exporter_registry: Dict[str, Type[Exporter]] = {}
        for exporter in exporters:
            self.register_exporter(exporter)
        self._objects = {}
        self.ignore_unknown = ignore_unknown

    def register_exporter(self, exporter_cls):
        kind = ID.kind_for_model(exporter_cls.Meta.model)
        self.__exporter_registry[kind] = exporter_cls

    def dump_objects(self):
        for obj in self._objects.values():
            print(f'* {obj.id}')
            print('\n  Fields:')
            for k, v in obj.serialized_data.items():
                print(f'  - {k} = {repr(v)}')
            if obj.attachments:
                print('\n  Attachments:')
                for a in obj.attachments:
                    print(f'  - {a.id}: {repr(a.key)}')
            print()

    def _exporter_for_model(self, obj: Model, raise_exception=True):
        return self._exporter_for_kind(ID.kind_for_model(obj), raise_exception=raise_exception)

    def _exporter_for_kind(self, kind: str, raise_exception=True):
        if kind not in self.__exporter_registry:
            if not raise_exception:
                return None
            raise ModelClassNotRegistered(kind)
        return self.__exporter_registry[kind]



from haul.policy import ImportPolicy, RelinkAction
from io import BytesIO
from test_app.app import reset
from test_app.models import Book, Author, Tag
from test_app.export import BookExporter, AuthorExporter, TagExporter

from haul import ImportContainer, ExportContainer


EXPORTERS = [BookExporter, AuthorExporter, TagExporter]


def test_discard_chain():
    reset()
    a1 = Author.objects.create(name='1')
    Author.objects.create(name='2')
    Book.objects.create(name='b1', author=a1)

    d = BytesIO()

    c = ExportContainer(exporters=EXPORTERS)
    c.export_objects(Author.objects.all())
    c.export_objects(Book.objects.all())
    c.write(d)

    reset()

    class Policy(ImportPolicy):
        def relink_object(self, model_cls, obj):
            if model_cls is Author and obj.fields['name'] == '1':
                return RelinkAction.Discard()
            return RelinkAction.Create()

    c = ImportContainer(exporters=EXPORTERS, policy=Policy())
    with c.read(d):
        c.import_objects()

    assert not Author.objects.filter(name='1').exists()
    assert Author.objects.filter(name='2').exists()
    assert Book.objects.count() == 0


def test_discard_optional():
    reset()
    a1 = Author.objects.create(name='1')
    a2 = Author.objects.create(name='2')
    Book.objects.create(name='b1', author=a1, coauthor=a2)

    d = BytesIO()

    c = ExportContainer(exporters=EXPORTERS)
    c.export_objects(Book.objects.all())
    c.write(d)

    reset()

    class Policy(ImportPolicy):
        def relink_object(self, model_cls, obj):
            if model_cls is Author and obj.fields['name'] == '2':
                return RelinkAction.Discard()
            return RelinkAction.Create()

    c = ImportContainer(exporters=EXPORTERS, policy=Policy())
    with c.read(d):
        c.import_objects()

    assert Author.objects.count() == 1
    assert Book.objects.count() == 1
    assert Book.objects.first().coauthor is None


def test_relink_fields():
    reset()
    a1 = Author.objects.create(name='1')
    Author.objects.create(name='2')
    Book.objects.create(name='b1', author=a1)

    d = BytesIO()

    c = ExportContainer(exporters=EXPORTERS)
    c.export_objects(Author.objects.all())
    c.export_objects(Book.objects.all())
    c.write(d)

    Book.objects.all().delete()
    Author.objects.filter(name='2').delete()

    class Policy(ImportPolicy):
        def relink_object(self, model_cls, obj):
            if model_cls is Author:
                return RelinkAction.LinkByFields(
                    lookup_fields=('name',),
                    fallback=RelinkAction.Create(),
                )
            return RelinkAction.Create()

    c = ImportContainer(exporters=EXPORTERS, policy=Policy())
    with c.read(d):
        c.import_objects()

    assert Author.objects.count() == 2
    assert Book.objects.count() == 1
    assert Book.objects.first().author == Author.objects.get(name='1')


def test_relink_and_overwrite_fields():
    reset()
    a = Author.objects.create(name='a')
    Book.objects.create(name='b1', isbn='101', author=a)
    Book.objects.create(name='b2', isbn='102', author=a)

    d = BytesIO()

    c = ExportContainer(exporters=EXPORTERS)
    c.export_objects(Book.objects.all())
    c.write(d)

    Book.objects.filter(name='b2').update(name='updated')

    class Policy(ImportPolicy):
        def relink_object(self, model_cls, obj):
            if model_cls is Book:
                return RelinkAction.LinkByFields(
                    lookup_fields=('isbn',),
                    overwrite_fields=('name',),
                )
            if model_cls is Author:
                return RelinkAction.LinkByFields(lookup_fields=('name',))

    c = ImportContainer(exporters=EXPORTERS, policy=Policy())
    with c.read(d):
        c.import_objects()

    assert Book.objects.count() == 2
    assert Book.objects.filter(name='b2').exists()
    assert not Book.objects.filter(name='updated').exists()


def test_relink_by_pk():
    reset()
    a = Author.objects.create(name='a')
    Book.objects.create(name='b1', author=a)
    Book.objects.create(name='b2', author=a)

    d = BytesIO()

    c = ExportContainer(exporters=EXPORTERS)
    c.export_objects(Book.objects.all())
    c.write(d)

    Book.objects.filter(name='b2').update(name='updated')

    class Policy(ImportPolicy):
        def relink_object(self, model_cls, obj):
            return RelinkAction.LinkByPK()

    c = ImportContainer(exporters=EXPORTERS, policy=Policy())
    with c.read(d):
        c.import_objects()

    assert Author.objects.count() == 1
    assert Book.objects.count() == 2
    assert Book.objects.filter(name='updated').exists()


def test_relink_to_instance():
    reset()
    a = Author.objects.create(name='a')
    Book.objects.create(name='b1', author=a)
    Book.objects.create(name='b2', author=a)

    d = BytesIO()

    c = ExportContainer(exporters=EXPORTERS)
    c.export_objects(Book.objects.all())
    c.write(d)

    imp_a = Author.objects.create(name='imported books author')

    class Policy(ImportPolicy):
        def relink_object(self, model_cls, obj):
            if model_cls is Author:
                return RelinkAction.LinkToInstance(pk=imp_a.pk)
            return RelinkAction.Create()

    c = ImportContainer(exporters=EXPORTERS, policy=Policy())
    with c.read(d):
        c.import_objects()

    assert Author.objects.count() == 2
    assert Book.objects.count() == 4
    assert imp_a.books.count() == 2



import logging
from dataclasses import dataclass, field
from enum import Enum, auto
from django.db.models import Model
from typing import Dict, Hashable, List, Sequence, Tuple, Type, Optional, Any, IO, Union
from typing_extensions import Literal
from .types import ID, Attachment, ObjectData
from ._util import get_model_options

logger = logging.getLogger('haul.policy')

RelinkReturnType = Union[Model, Literal[False]]


class ManyToManyImportAction(Enum):
    REPLACE = auto()
    APPEND = auto()


@dataclass(frozen=True)
class BaseRelinkAction:
    '''
    Base class for new relink actions.

    Override :func:`execute` to return a list of relinked model instances, ``None``
    or a literal ``False`` to discard the object.
    '''

    fallback: Optional['BaseRelinkAction'] = None

    def execute(self, model_cls: Type[Model], objects: List[ObjectData], policy: 'ImportPolicy') -> Sequence[Optional[RelinkReturnType]]:
        raise NotImplementedError

    def _execute(self, model_cls: Type[Model], objects: List[ObjectData], policy: 'ImportPolicy') -> Sequence[RelinkReturnType]:
        instances = list(self.execute(model_cls, objects, policy))
        if any(x is None for x in instances):
            fallback_objects, fallback_indices = zip(*[
                (objects[i], i)
                for (i, x) in enumerate(instances)
                if x is None
            ])
            if len(fallback_indices):
                if not self.fallback:
                    raise RuntimeError(f'{self} failed to resolve some items and there was no fallback provided')
                logger.debug(f'Fallback to {self.fallback} for {len(fallback_objects)} objects')
                fallback_instances = self.fallback.execute(model_cls, fallback_objects, policy)  # type: ignore
                for instance, index in zip(fallback_instances, fallback_indices):
                    instances[index] = instance
        return instances  # type: ignore


@dataclass(frozen=True)
class BaseRelinkActionWithFieldsOverwrite(BaseRelinkAction):
    '''
    A base relink action class that automatically overwrites model fields
    set via :attr:`overwrite_fields`.
    '''

    overwrite_fields: Tuple[str, ...] = tuple()

    def _execute(self, model_cls: Type[Model], objects: List[ObjectData], policy: 'ImportPolicy') -> Sequence[RelinkReturnType]:
        instances = super()._execute(model_cls, objects, policy)
        if self.overwrite_fields:
            for obj, instance in zip(objects, instances):
                if instance and instance is not False:
                    assert obj.fields is not None

                    if self.overwrite_fields == ('__all__',):
                        fields = {k: v for k, v in obj.fields.items() if k not in ['pk', 'id']}
                    else:
                        fields = {k: v for k, v in obj.fields.items() if k in self.overwrite_fields}
                    _assign_model_fields(instance, fields, policy)
                    instance.save()

        return instances


class RelinkAction:
    @dataclass(frozen=True)
    class LogToConsole(BaseRelinkAction):
        '''
        A relink action that prints the objects and links them to ``None``.
        '''
        def execute(self, model_cls: Type[Model], objects: List[ObjectData], policy: 'ImportPolicy') -> Sequence[Optional[RelinkReturnType]]:
            for obj in objects:
                assert obj.fields is not None
                logger.info(f'[{obj.id}] -> {model_cls(**obj.fields)}')
            return [None] * len(objects)

    @dataclass(frozen=True)
    class Create(BaseRelinkAction):
        '''
        A relink action that unconditionally creates new objects.
        '''

        #: A list of fields to ignore during import
        ignore_fields: Tuple[str, ...] = field(default_factory=tuple)

        def __post_init__(self):
            if not isinstance(self.ignore_fields, tuple):
                raise ValueError('ignore_fields must be a tuple')

        def execute(self, model_cls: Type[Model], objects: List[ObjectData], policy: 'ImportPolicy') -> List[RelinkReturnType]:
            fieldsets = [
                {k: v for k, v in obj.fields.items() if k not in self.ignore_fields}
                for obj in objects
                if obj.fields
            ]
            if False:  # TODO detect mysql
                instances = model_cls.objects.bulk_create([
                    model_cls(**_get_non_x2m_fields(model_cls, fields))
                    for fields in fieldsets
                ])
            else:
                instances = [
                    model_cls.objects.create(**_get_non_x2m_fields(model_cls, fields))
                    for fields in fieldsets
                ]
            for instance, fields in zip(instances, fieldsets):
                _assign_model_fields(instance, fields, policy)

            return instances

    @dataclass(frozen=True)
    class Discard(BaseRelinkAction):
        '''
        A relink action that unconditionally discards all objects.
        '''
        def execute(self, _model_cls, objects: List[ObjectData], policy: 'ImportPolicy') -> Sequence[Optional[RelinkReturnType]]:
            return [False] * len(objects)

    @dataclass(frozen=True)
    class LinkByFields(BaseRelinkActionWithFieldsOverwrite):
        '''
        A relink action that tries to look up existing objects by a set of fields.
        '''

        #: A list of fields to use to look up existing objects
        lookup_fields: Tuple[str, ...] = field(default_factory=tuple)

        def __post_init__(self):
            if not isinstance(self.lookup_fields, tuple):
                raise ValueError('lookup_fields must be a tuple')
            if not len(self.lookup_fields):
                raise ValueError('lookup_fields must be set')
            if not isinstance(self.overwrite_fields, tuple):
                raise ValueError('overwrite_fields must be a tuple')

        def execute(self, model_cls: Type[Model], objects: List[ObjectData], policy: 'ImportPolicy') -> Sequence[Optional[RelinkReturnType]]:
            search_cache = {}
            instances = []
            for obj in objects:
                assert obj.fields is not None
                values = tuple(obj.fields.get(k) for k in self.lookup_fields)
                if values not in search_cache:
                    query = dict(zip(self.lookup_fields, values))
                    found = model_cls.objects.filter(**query).first()
                    search_cache[values] = found

                instances.append(search_cache.get(values))

            return instances

    @dataclass(frozen=True)
    class LinkByPK(BaseRelinkActionWithFieldsOverwrite):
        '''
        A relink action that tries to map objects by their literal PK value.

        This only works if the PKs are same in the source and destination databases.
        '''

        def __post_init__(self):
            if not isinstance(self.overwrite_fields, tuple):
                raise ValueError('overwrite_fields must be a tuple')

        def execute(self, model_cls: Type[Model], objects: List[ObjectData], policy: 'ImportPolicy') -> List[Optional[RelinkReturnType]]:
            pks = [obj.id.pk for obj in objects]
            search_cache = {
                x.pk: x
                for x in model_cls.objects.filter(pk__in=pks)
            }
            return [
                search_cache.get(obj.id.pk)
                for obj in objects
            ]

    @dataclass(frozen=True)
    class LinkToInstance(BaseRelinkActionWithFieldsOverwrite):
        '''
        A relink action that links the object to a specific existing object.
        '''

        #: PK of the existing object to link to
        pk: Hashable = None

        def __post_init__(self):
            if not self.pk:
                raise ValueError('pk must be set')

        def execute(self, model_cls: Type[Model], objects: List[ObjectData], policy: 'ImportPolicy') -> Sequence[RelinkReturnType]:
            instance = model_cls.objects.get(pk=self.pk)
            return [instance] * len(objects)


class ImportPolicy:
    '''
    A behaviour class that defines how to handle objects during import.
    Override its methods and pass it to :class:`ImportContainer` to customize the behaviour.
    '''

    def preprocess_object_fields(self, model_cls: Type[Model], fields: Dict[str, Any]):
        '''
        Called after objects have been deserialized.
        You can modify the ``fields`` dict in place here.
        '''

    def postprocess_object_fields(self, model_cls: Type[Model], fields: Dict[str, Any]):
        '''
        Called after :func:`relink_object` and before the relink actions have been applied.
        You can modify the ``fields`` dict in place here.
        '''

    def relink_object(self, model_cls: Type[Model], obj: ObjectData) -> BaseRelinkAction:
        '''
        Called to decide how the imported objects should be mapped onto existing objects.
        This method should return an instance of a :class:`BaseRelinkAction` subclass.

        The default implementation just returns ``RelinkAction.Create()``
        '''
        return RelinkAction.Create()

    def process_many_to_many(self, instance: Model, field: str) -> ManyToManyImportAction:
        '''
        Called for each M2M relation on a model. You can either return a ``ManyToManyImportAction.REPLACE`` or ``ManyToManyImportAction.APPEND`` here. The default is ``REPLACE``.
        '''
        return ManyToManyImportAction.REPLACE

    def process_attachment(self, instance: Model, key: Any, stream: IO[bytes]):
        '''
        Called for each attachment on a model.
        This is where you read and process the attachment contents.
        Note that the stream is not guaranteed to be open after this method exits
        and is guaranteed to be closed once the :func:`ImportContainer.import` context exits.
        '''
        raise NotImplementedError('Attachments found but process_attachment is not implemented in the import policy')

    def post_object_import(self, instance: Model):
        pass


class ExportPolicy:
    '''
    A behaviour class that defines how to handle objects during export.
    '''

    def get_attachments(self, instance: Model) -> List[Attachment]:
        '''
        Override to add file or stream based attachments to an object. See :class:`Attachment`.
        '''
        return []

    def should_export_object(self, instance: Model):
        '''
        Override to additionally filter objects during export.
        '''
        return True

    def should_follow_reference(self, instance: Model, target: ID, field: str):
        '''
        Override to additionally filter FK references during export.
        '''
        return True

    # def postprocess_object(self, instance: Model, data: Mapping[str, Any]):


def _get_non_x2m_fields(model_cls: Type[Model], fields: Dict[str, Any]):
    model_meta = get_model_options(model_cls)
    return {
        k: v
        for k, v in fields.items()
        if not model_meta.get_field(k).many_to_many and
        not model_meta.get_field(k).one_to_many
    }


def _assign_model_fields(instance: Model, fields: Dict[str, Any], policy: ImportPolicy):
    model_meta = get_model_options(instance.__class__)
    for key, value in fields.items():
        field_meta = model_meta.get_field(key)
        if field_meta.many_to_many:
            action = policy.process_many_to_many(instance, key)
            if action == ManyToManyImportAction.REPLACE:
                getattr(instance, key).set(value)
            else:
                for obj in value:
                    getattr(instance, key).add(obj)
        elif field_meta.one_to_many:
            # We don't need to actually *set* the field
            pass
        else:
            setattr(instance, key, value)
    instance.save()



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = [
    'django>=3',
    'djangorestframework>=3',
    'ruamel.yaml>=0.17',
    'graphlib-backport; python_version<"3.9"',
]

test_requirements = ['pytest>=3', 'typing_extensions']

setup(
    author="Eugene Pankov`",
    author_email='e.pankov@elements.tv',
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Structured object import/export for Django ORM",
    install_requires=requirements,
    license="MIT",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords=['django', 'orm', 'export', 'import'],
    name='django-haul',
    packages=find_packages(include=['haul', 'haul.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/elements-storage/haul',
    version='0.0.10',
    zip_safe=False,
)
